/*
 *	use_any_pkg: let's try to use any packages named on the command line!
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>


//
// void *p = lookup_function( void *dl, char *module, char *funcname );
//	look within dl, a dynamic library opened by dlopen(), for
//	either the global symbol <funcname> or the global symbol
//	<module_funcname>.
//	If we find either, return a pointer to the first one we found.
//	If not, return NULL.
//
void *lookup_function( void *dl, char *module, char *funcname )
{
	// can we look up the symbol funcname inside the dl?
	void *p = dlsym( dl, funcname );
	if( p != NULL )
	{
		return p;
	}

	//printf( "Not found %s in %s\n", funcname, module );

	// can we look up the module-qualified symbol inside the dl?
	char fullname[1024];
	sprintf( fullname, "%s_%s", module, funcname );
	p = dlsym( dl, fullname );
	if( p != NULL )
	{
		return p;
	}
	//printf( "Not found %s in %s either\n", fullname, module );
	return NULL;
}


// void_void_f: a pointer to a void->void function
typedef void (*void_void_f)( void );


// int_void_f: a pointer to a void->int function
typedef int (*int_void_f)( void );


//
// f1_f2( char *module );
//	load the given module via libdl, and try to
//	find functions f1 and f2 in that module,
//	invoking both of them if you find them.
//
void f1_f2( char *module )
{
	char libname[1024];
	sprintf( libname, "lib%s.so", module );

	// open the shared library
	void *dl = dlopen( libname, RTLD_NOW );
	if( dl == NULL )
	{
		fprintf( stderr, "Can't dlopen %s\n", libname );
		return;
	}

	// Can we find (void->void) function f1 in module?
	void_void_f f1 = lookup_function( dl, module, "f1" );
	if( f1 == NULL )
	{
		fprintf( stderr, "Failed to find f1 or %s_f1 in %s\n",
			module, libname );
		return;
	}

	// Can we find (void->int) function f2 in module..
	int_void_f f2 = lookup_function( dl, module, "f2" );
	if( f2 == NULL )
	{
		fprintf( stderr, "Failed to find f2 or %s_f2 in %s\n",
			module, libname );
		return;
	}

	// ok, now we have f1, a pointer to the module's f1 or
	// module_f1 function. let's just call it:
	printf( "Calling %s's f1 function\n", module );
	(*f1)();

	// ditto, now we have f2, let's just call it:
	int n = (*f2)();
	printf( "Called %s's f2 function, result is %d\n", module, n );
}


int main( int argc, char **argv )
{
	for( int i=1; argv[i] != NULL; i++ )
	{
		f1_f2( argv[i] );
		printf( "\n" );
	}
	return 0;
}
